"""demo-01.py - 8x8 ground tiles, single screen, non-scrolling

This is a very basic demo. It uses a traditional game loop with synchronous update() and draw() calls.

Two layers for the ground graphics.
One layer for the mobile graphics.
Two layers for the overhead graphics.

All layers are combined into one composite buffer.

1 composite buffer
5 sprite layers
5 spatial hashes
1 on-screen list
"""

import cProfile
import os
import pstats
import random
import sys

import pygame
from pygame import Color, KEYDOWN, KEYUP, QUIT, K_UP, K_DOWN, K_LEFT, K_RIGHT, K_TAB, K_SPACE, K_ESCAPE

sys.path.insert(0, 'gamelib')

from compositebuffer import CompositeBuffer
from spatialhash import SpatialHash


if sys.version_info[0] == 3:
    xrange = range

RESOLUTION = 800, 600


class Sprite(object):
    def update(self, dt):
        pass


class Player(Sprite):
    def __init__(self, image, rect, x, y, world_rect):
        self.image = image
        self.rect = rect
        self.x = x
        self.y = y
        self.dx = 0
        self.dy = 0
        self.speed = 100
        self.world_rect = world_rect

    def update(self, dt):
        if self.dx or self.dy:
            x = self.x + self.dx * self.speed * dt
            y = self.y + self.dy * self.speed * dt
            self.rect.topleft = x, y
            if self.world_rect.contains(self.rect):
                self.x = x
                self.y = y
            else:
                self.rect.clamp_ip(self.world_rect)
                self.x, self.y = self.rect.topleft


class Game(object):
    # Tunable
    NUM_ROCKS = 100             # num rocks!
    NUM_TREES = 40              # num trees and shadows!
    COMPOSITE_TILE_SIZE = 64    # size for CompositeBuffer and SpatialHash
    MAX_FPS = 0                 # cap fps >= 0
    TIMER_RESET = 1.0           # seconds between window caption updates

    # Non-tunable
    GROUND, ON_GROUND, ABOVE_GROUND, SHADOWS, OVERHEAD = list(range(5))
    MODE_COMPOSITE, MODE_SPRITES = 0, 1

    def __init__(self):
        self.screen = pygame.display.get_surface()
        self.screen_rect = self.screen.get_rect()
        self.clear_color = Color('black')

        self.clock = pygame.time.Clock()
        self.max_fps = self.MAX_FPS
        self.dt = 1.0 / self.max_fps if self.max_fps > 0 else 0
        self.timer_reset = self.TIMER_RESET
        self.timer = self.timer_reset

        self.player = None
        self.camera = self.screen_rect.copy()

        self.composite_buffer = CompositeBuffer(tile_size=self.COMPOSITE_TILE_SIZE)
        # Maintain a couple of nested structures for speed:
        # sprites for update and rendering; mode layers for compositing
        self.sprites = [
            [],   # 0: ground
            [],   # 1: on ground
            [],   # 2: above ground
            [],   # 3: shadows
            [],   # 4: overhead
        ]
        self.composite_hash = [
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 0: ground
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 1: on ground
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 2: above ground
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 3: shadows
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 4: overhead
        ]
        self.sprite_getters = [h.intersect_entities for h in self.composite_hash]
        self.on_screen = []
        self.mode = self.MODE_SPRITES

        self.make_sprites()

        self.running = False

    def make_sprites(self):
        # Clear everything to make this reentrant.
        for sprites in self.sprites:
            del sprites[:]
        for layer in self.composite_hash:
            layer.clear()
        self.composite_buffer.clear()
        del self.on_screen[:]

        # Dirt - make the background tiles
        font = pygame.font.SysFont('sans', 11)
        sprites = []
        size = 16
        label = False
        y_range = xrange(self.screen_rect.y, self.screen_rect.bottom + size, size)
        for x in xrange(self.screen_rect.x, self.screen_rect.right + size, size):
            for y in y_range:
                s = Sprite()
                s.idx = x, y
                s.image = pygame.Surface((size, size))
                s.rect = s.image.get_rect(topleft=(x, y))
                s.image.fill(Color('tan1'))
                pos_text = '{},{}'.format(x, y)
                if label and font.size(pos_text)[0] < size:
                    pos_label = font.render(pos_text, True, Color('black'))
                    s.image.blit(pos_label, (1, 1))
                pygame.draw.rect(s.image, Color('tan2'), s.image.get_rect(), 1)
                sprites.append(s)
        self.sprites[self.GROUND].extend(sprites)
        self.composite_hash[self.GROUND].addlist(sprites)

        # Rocks
        sprites = []
        for i in range(self.NUM_ROCKS):
            x = random.randrange(self.screen_rect.x, self.screen_rect.right)
            y = random.randrange(self.screen_rect.y, self.screen_rect.bottom)
            size = random.randrange(15, 25)
            r = size // 2
            s = Sprite()
            s.idx = i
            s.image = pygame.Surface((size, size))
            s.rect = s.image.get_rect(topleft=(x, y)).clamp(self.screen_rect)
            s.image.set_colorkey(Color('black'))
            pygame.draw.circle(s.image, Color('grey50'), (r, r), r)
            pygame.draw.circle(s.image, Color('grey30'), (r, r), r, 1)
            sprites.append(s)
        self.sprites[self.ON_GROUND].extend(sprites)
        self.composite_hash[self.ON_GROUND].addlist(sprites)

        # Player
        sprites = []
        size = 40
        r = size // 2
        x = random.randrange(self.screen_rect.x, self.screen_rect.right)
        y = random.randrange(self.screen_rect.y, self.screen_rect.bottom)
        image = pygame.Surface((size, size))
        rect = image.get_rect(topleft=(x, y)).clamp(self.screen_rect)
        image.set_colorkey(Color('black'))
        pygame.draw.circle(image, Color('skyblue2'), (r, r), r)
        pygame.draw.circle(image, Color('skyblue4'), (r, r), r, 1)
        x = rect.x
        y = rect.y
        s = Player(image, rect, x, y, self.screen_rect)
        s.idx = 0
        sprites.append(s)
        self.sprites[self.ABOVE_GROUND].extend(sprites)
        self.composite_hash[self.ABOVE_GROUND].addlist(sprites)
        self.player = s

        # Trees and shadows
        s_sprites = []
        sh_sprites = []
        for i in xrange(self.NUM_TREES):
            x = random.randrange(self.screen_rect.x, self.screen_rect.right)
            y = random.randrange(self.screen_rect.y, self.screen_rect.bottom)
            size = random.randrange(54, 75)
            r = size // 2
            s = Sprite()
            s.idx = i
            s.image = pygame.Surface((size, size))
            s.rect = s.image.get_rect(center=(x, y)).clamp(self.screen_rect)
            s.image.set_colorkey(Color('black'))
            pygame.draw.circle(s.image, Color('green2'), (r, r), r)
            pygame.draw.circle(s.image, Color('green4'), (r, r), r, 1)
            sh = Sprite()
            sh.image = s.image.copy().convert_alpha()
            sh.rect = s.rect.move(45, 30)
            pygame.draw.circle(sh.image, Color('grey20'), (r, r), r)
            pygame.draw.circle(sh.image, Color('black'), (r, r), r, 1)
            sh.image.fill((255, 255, 255, 32), None, pygame.BLEND_RGBA_MULT)
            s_sprites.append(s)
            sh_sprites.append(sh)
        self.sprites[self.SHADOWS].extend(sh_sprites)
        self.sprites[self.OVERHEAD].extend(s_sprites)
        self.composite_hash[self.SHADOWS].addlist(sh_sprites)
        self.composite_hash[self.OVERHEAD].addlist(s_sprites)

        # Sort the sprites so they render in the same order as the composite tiles.
        for i, sprites in enumerate(self.sprites):
            sprites.sort(key=self.composite_buffer.sort_key)

    def run(self):
        self.running = True
        while self.running:
            self.update_timer()
            self.handle_events()
            self.update(self.dt)
            self.draw()

    def switch_mode(self):
        self.mode = self.MODE_COMPOSITE if self.mode == self.MODE_SPRITES else self.MODE_SPRITES
        if self.mode == self.MODE_COMPOSITE:
            self.composite_buffer.drain()
            self.composite_hash[self.ABOVE_GROUND].add(self.player)

    def update(self, dt):
        if self.mode == self.MODE_COMPOSITE:
            self.update_composite(dt)
        else:
            self.update_sprites(dt)

    def update_timer(self):
        self.dt = self.clock.tick(self.max_fps) / 1000.0
        self.timer -= self.dt
        if self.timer <= 0.0:
            self.timer = self.timer_reset
            mode = ('Composite', 'Sprites')
            ground = len(self.sprites[self.GROUND])
            rocks = len(self.sprites[self.ON_GROUND])
            trees = len(self.sprites[self.OVERHEAD]) * 2
            composites = len(self.on_screen)
            things = 'Ground {}, Rocks {}, Trees+Shadows {}, Composites {}'.format(ground, rocks, trees, composites)
            pygame.display.set_caption('TAB: {}> | {} | FPS {}'.format(
                mode[self.mode], things, int(self.clock.get_fps())))

    def update_composite(self, dt):
        dirty_rects = []
        dirty = False
        if self.player.dx or self.player.dy:
            dirty = True
            dirty_rects.append(self.player.rect)
        for s in self.sprites[self.ABOVE_GROUND]:
            s.update(dt)
        if dirty:
            dirty_rects.append(self.player.rect)
            self.composite_buffer.set_dirty(dirty_rects)
        self.composite_hash[self.ABOVE_GROUND].addlist(self.sprites[self.ABOVE_GROUND])
        self.on_screen = self.composite_buffer.get_tiles(self.camera, self.sprite_getters)

    def update_sprites(self, dt):
        for s in self.sprites[self.ABOVE_GROUND]:
            s.update(dt)

    def draw(self):
        self.screen.fill(self.clear_color)
        if self.mode == self.MODE_COMPOSITE:
            self.draw_composite()
        else:
            self.draw_sprites()
        pygame.display.flip()

    def draw_composite(self):
        blit = self.screen.blit
        cx, cy = self.camera.topleft
        for s in self.on_screen:
            blit(s.image, s.rect.move(-cx, -cy))

    def draw_sprites(self):
        blit = self.screen.blit
        cx, cy = self.camera.topleft
        for layer in self.sprites:
            for s in layer:
                blit(s.image, s.rect.move(-cx, -cy))

    def handle_events(self):
        for e in pygame.event.get():
            if e.type == KEYDOWN:
                if e.key == K_UP:
                    self.player.dy -= 1
                elif e.key == K_DOWN:
                    self.player.dy += 1
                elif e.key == K_LEFT:
                    self.player.dx -= 1
                elif e.key == K_RIGHT:
                    self.player.dx += 1
                elif e.key == K_ESCAPE:
                    self.running = False
                elif e.key == K_TAB:
                    self.switch_mode()
                elif e.key == K_SPACE:
                    self.make_sprites()
            elif e.type == KEYUP:
                if e.key == K_UP:
                    self.player.dy += 1
                elif e.key == K_DOWN:
                    self.player.dy -= 1
                elif e.key == K_LEFT:
                    self.player.dx += 1
                elif e.key == K_RIGHT:
                    self.player.dx -= 1
            elif e.type == QUIT:
                self.running = False


if __name__ == '__main__':
    def main():
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        pygame.init()
        pygame.display.set_mode(RESOLUTION)
        Game().run()

    if False:
        cProfile.runctx('main()', globals(), locals(), 'prof.dat')
        p = pstats.Stats('prof.dat')
        p.sort_stats('time').print_stats()
    else:
        main()
