"""demo-02.py - 16x16 ground tiles, large map, scrolling

Two layers for the ground graphics.
One layer for the mobile graphics.
Two layers for the overhead graphics.

All layers are combined into one composite buffer.

1 composite buffer
5 sprite layers
5 spatial hashes
1 on-screen list
"""

import cProfile
import os
import pstats
import random
import sys

import pygame
from pygame import Color, KEYDOWN, KEYUP, QUIT, K_UP, K_DOWN, K_LEFT, K_RIGHT, K_TAB, K_SPACE, K_ESCAPE

sys.path.insert(0, 'gamelib')

from compositebuffer import CompositeBuffer
from spatialhash import SpatialHash
import hudlight
import hudthemes


if sys.version_info[0] == 3:
    xrange = range

MAX_FPS = 512
RESOLUTION = 800, 600
WORLD_SIZE = 4000, 4000
WORLD_RECT = -WORLD_SIZE[0] // 2, -WORLD_SIZE[1] // 2, WORLD_SIZE[0], WORLD_SIZE[1]
WORLD_BOUNDS = -WORLD_SIZE[0] // 2, -WORLD_SIZE[1] // 2, WORLD_SIZE[0] // 2, WORLD_SIZE[1] // 2


LAYERS = (
    DIRT_LAYER,
    ROCK_LAYER,
    LADY_LAYER,
    PLAYER_LAYER,
    TREESHADOW_LAYER,
    TREE_LAYER,
) = range(6)


class Sprite(object):
    layer = 0

    def __init__(self, image, rect, x, y, world_rect):
        self.image = image
        self.rect = rect
        self.x = x
        self.y = y
        self.world_rect = world_rect

    @staticmethod
    def sort_key(sprite):
        r = sprite.rect
        # rect.bottom, rect.x
        return sprite.layer, r[1] + r[3], r[0]

    def update(self, dt):
        pass


class Dirt(Sprite):
    layer = DIRT_LAYER


class Rock(Sprite):
    layer = ROCK_LAYER


class Player(Sprite):
    layer = PLAYER_LAYER

    def __init__(self, image, rect, x, y, world_rect):
        super(Player, self).__init__(image, rect, x, y, world_rect)
        self.dx = 0
        self.dy = 0
        self.speed = 100

    def update(self, dt):
        if self.dx or self.dy:
            x = self.x + self.dx * self.speed * dt
            y = self.y + self.dy * self.speed * dt
            self.rect.topleft = x, y
            if self.world_rect.contains(self.rect):
                self.x = x
                self.y = y
            else:
                self.rect.clamp_ip(self.world_rect)
                self.x, self.y = self.rect.topleft


class LadyFriend(Player):
    layer = LADY_LAYER

    def __init__(self, image, rect, x, y, world_rect):
        super(LadyFriend, self).__init__(image, rect, x, y, world_rect)
        self.dx = random.choice((-1, 1))
        self.dy = random.choice((-1, 1))

    def update(self, dt):
        x = self.x
        y = self.y
        new_x = x + self.dx * self.speed * dt
        new_y = y + self.dy * self.speed * dt
        self.rect.topleft = new_x, new_y
        if not self.world_rect.contains(self.rect):
            test_rect = self.rect.clamp(self.world_rect)
            if test_rect.x != self.rect.x:
                # self.rect.x = new_x = x
                self.dx = -self.dx
            if test_rect.y != self.rect.y:
                # self.rect.y = new_y = y
                self.dy = -self.dy
        self.x, self.y = new_x, new_y


class TreeShade(Sprite):
    layer = TREESHADOW_LAYER


class Tree(Sprite):
    layer = TREE_LAYER


class Game(object):
    # Tunable
    BACKGROUND_TILE_SIZE = 16   # background tiles will be square of this size
    NUM_ROCKS = 800             # num rocks!
    NUM_TREES = 500             # num trees and shadows!
    NUM_LADY_FRIENDS = 50       # num lady friends running away
    COMPOSITE_TILE_SIZE = 64    # size for CompositeBuffer and SpatialHash
    TIMER_RESET = 0.5           # seconds between window caption and HUD updates

    # Non-tunable
    BACKGROUND, ON_GROUND, ABOVE_GROUND, SHADOWS, OVERHEAD = list(range(5))
    MODE_COMPOSITE, MODE_SPRITES = 0, 1

    def __init__(self):
        self.screen = pygame.display.get_surface()
        self.screen_rect = self.screen.get_rect()
        self.clear_color = Color('black')

        self.clock = pygame.time.Clock()
        self.max_fps = MAX_FPS
        self.dt = 1.0 / self.max_fps if self.max_fps > 0 else 0
        self.timer_reset = self.TIMER_RESET
        self.timer = self.timer_reset

        self.world_rect = pygame.Rect(WORLD_RECT)
        self.player = None
        self.camera = pygame.Rect(self.screen_rect)

        self.composite_buffer = CompositeBuffer(tile_size=self.COMPOSITE_TILE_SIZE, sort_key=Sprite.sort_key)
        # Maintain a couple of nested structures for speed:
        # sprites for update and rendering; mode layers for compositing
        self.sprites = [
            [],   # 0: ground
            [],   # 1: on ground
            [],   # 2: above ground
            [],   # 3: shadows
            [],   # 4: overhead
        ]
        self.composite_src_layers = [
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 0: ground
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 1: on ground
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 2: above ground
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 3: shadows
            SpatialHash(self.COMPOSITE_TILE_SIZE),  # 4: overhead
        ]
        self.sprite_getters = [h.intersect_entities for h in self.composite_src_layers]
        self.on_screen = []
        self.mode = self.MODE_COMPOSITE

        self.frames = 1
        self.num_dirty = 0
        self.num_tiles_rendered = 0
        self.num_sprites_rendered = 0
        self.num_new = 0

        self.make_sprites()
        self.hud = self.make_hud()

        self.running = False

    def make_sprites(self):
        # Clear everything to make this reentrant.
        for sprites in self.sprites:
            del sprites[:]
        for layer in self.composite_src_layers:
            layer.clear()
        self.composite_buffer.clear()
        del self.on_screen[:]

        # Dirt - make the background tiles
        font = pygame.font.SysFont('sans', 11)
        sprites = []
        size = self.BACKGROUND_TILE_SIZE
        label = False
        total_tiles = (self.world_rect.height + size) * (self.world_rect.width + size) // size
        n = 0

        image = pygame.Surface((size, size))
        image.fill(Color('tan1'))
        pygame.draw.rect(image, Color('tan2'), image.get_rect(), 1)

        y_range = xrange(self.world_rect.y, self.world_rect.bottom + size, size)
        for x in xrange(self.world_rect.x, self.world_rect.right + size, size):
            n += 1
            pygame.display.set_caption('Making background tiles {} of {}'.format(n, total_tiles))
            for y in y_range:
                _image = image
                rect = _image.get_rect(topleft=(x, y))
                pos_text = '{},{}'.format(x, y)
                if label and font.size(pos_text)[0] < size:
                    _image = image.copy()
                    pos_label = font.render(pos_text, True, Color('black'))
                    _image.blit(pos_label, (1, 1))
                s = Dirt(_image, rect, x, y, self.world_rect)
                sprites.append(s)
        self.sprites[self.BACKGROUND].extend(sprites)
        self.composite_src_layers[self.BACKGROUND].addlist(sprites)

        # Rocks
        sprites = []
        total_tiles = self.NUM_ROCKS
        n = 0
        images = {}
        for i in range(self.NUM_ROCKS):
            n += 1
            pygame.display.set_caption('Making rock tiles {} of {}'.format(n, total_tiles))
            x = random.randrange(self.world_rect.x, self.world_rect.right)
            y = random.randrange(self.world_rect.y, self.world_rect.bottom)
            size = random.randrange(15, 25)
            r = size // 2
            if size in images:
                _image = images[size]
            else:
                _image = pygame.Surface((size, size))
                _image.set_colorkey(Color('black'))
                pygame.draw.circle(_image, Color('grey50'), (r, r), r)
                pygame.draw.circle(_image, Color('grey30'), (r, r), r, 1)
                images[size] = _image
            rect = _image.get_rect(topleft=(x, y)).clamp(self.world_rect)
            s = Rock(_image, rect, x, y, self.world_rect)
            sprites.append(s)
        self.sprites[self.ON_GROUND].extend(sprites)
        self.composite_src_layers[self.ON_GROUND].addlist(sprites)

        # Player
        sprites = []
        size = 40
        r = size // 2
        ppls_info = (
            (Player, 1, 'skyblue2', 'skyblue4'),
            (LadyFriend, self.NUM_LADY_FRIENDS, 'pink2', 'pink4'),
        )
        for Ppl, count, color1, color2 in ppls_info:
            color1 = pygame.Color(color1)
            color2 = pygame.Color(color2)
            image = pygame.Surface((size, size))
            image.set_colorkey(Color('black'))
            pygame.draw.circle(image, color1, (r, r), r)
            pygame.draw.circle(image, color2, (r, r), r, 1)
            for i in xrange(count):
                if Ppl is Player:
                    x, y = WORLD_BOUNDS[0], WORLD_BOUNDS[1]
                else:
                    x = random.randrange(self.world_rect.x, self.world_rect.right)
                    y = random.randrange(self.world_rect.y, self.world_rect.bottom)
                rect = image.get_rect(topleft=(x, y)).clamp(self.world_rect)
                x = rect.x
                y = rect.y
                s = Ppl(image, rect, x, y, self.world_rect)
                if Ppl is Player:
                    self.player = s
                sprites.append(s)
        self.sprites[self.ABOVE_GROUND].extend(sprites)
        self.composite_src_layers[self.ABOVE_GROUND].addlist(sprites)

        # Trees and shadows
        s_sprites = []
        sh_sprites = []
        total_tiles = self.NUM_TREES
        n = 0
        images = {}
        for i in xrange(self.NUM_TREES):
            n += 1
            pygame.display.set_caption('Making trees tiles {} of {}'.format(n, total_tiles))
            x = random.randrange(self.world_rect.x, self.world_rect.right)
            y = random.randrange(self.world_rect.y, self.world_rect.bottom)
            size = random.randrange(54, 75)
            r = size // 2
            if size in images:
                s_image, sh_image = images[size]
            else:
                s_image = pygame.Surface((size, size))
                s_image.set_colorkey(Color('black'))
                pygame.draw.circle(s_image, Color('green2'), (r, r), r)
                pygame.draw.circle(s_image, Color('green4'), (r, r), r, 1)
                sh_image = s_image.copy().convert_alpha()
                pygame.draw.circle(sh_image, Color('grey20'), (r, r), r)
                pygame.draw.circle(sh_image, Color('black'), (r, r), r, 1)
                sh_image.fill((255, 255, 255, 32), None, pygame.BLEND_RGBA_MULT)
                images[size] = s_image, sh_image
            s_rect = s_image.get_rect(center=(x, y)).clamp(self.world_rect)
            sh_rect = s_rect.move(45, 30)
            s = Tree(s_image, s_rect, x, y, self.world_rect)
            sh = TreeShade(sh_image, sh_rect, sh_rect.x, sh_rect.y, self.world_rect)
            s_sprites.append(s)
            sh_sprites.append(sh)
        self.sprites[self.SHADOWS].extend(sh_sprites)
        self.sprites[self.OVERHEAD].extend(s_sprites)
        self.composite_src_layers[self.SHADOWS].addlist(sh_sprites)
        self.composite_src_layers[self.OVERHEAD].addlist(s_sprites)

        # Sort the sprites so they render in the same order as the composite tiles.
        for i, sprites in enumerate(self.sprites):
            sprites.sort(key=Sprite.sort_key)

    def make_hud(self):
        def get_player():
            return self.player.rect.center

        def get_on_screen():
            return len(self.on_screen)

        def get_tiles():
            return len(self.composite_buffer)

        def get_cache():
            bytesize = self.screen.get_bytesize()
            num_cached = self.composite_buffer.num_cached
            cache_size = float(self.composite_buffer.cache_size * bytesize) / 2 ** 20
            cache_max = float(self.composite_buffer.cache_max_px * bytesize) / 2 ** 20
            return num_cached, cache_size, cache_max

        def get_comp_stats():
            num_dirty = self.num_dirty
            num_tiles_rendered = self.num_tiles_rendered
            frames = self.frames
            return int(num_dirty / frames), int(num_tiles_rendered / frames)

        def get_sprites():
            num_new = self.num_new
            num_sprites_rendered = self.num_sprites_rendered
            frames = self.frames
            return num_new, int(num_sprites_rendered / frames)

        hudlight.set_font_template('data/font/%s')
        theme = hudthemes.HUD_THEMES['verabd'].copy()
        theme.update({'fontsize': 18, 'color': 'lightblue1', 'gcolor': 'lightblue2', 'ocolor': 'black', 'owidth': 0.75})
        hud = hudlight.HUD(**theme)
        hud.add('fps', 'FPS {:0.0f}', self.clock.get_fps(), callback=self.clock.get_fps)
        hud.add('world', 'World ({}, {}) - ({}, {})', *WORLD_BOUNDS)
        hud.add('player', 'Player {}, {}', *get_player(), callback=get_player)
        hud.add('on_screen', 'On Screen {}', get_on_screen(), callback=get_on_screen)
        hud.add('comp tiles', 'Comp Tiles {}', get_tiles(), callback=get_tiles)
        hud.add('comp cache', 'Comp Cache {} tiles, {:0.1f} / {:0.1f} MB', *get_cache(), callback=get_cache)
        hud.add('comp dirty', 'Comp Dirty {}/tick, Rendered {}/tick', *get_comp_stats(), callback=get_comp_stats)
        hud.add('comp sprites', 'Comp Sprites New {}, Rendered {}/tick', *get_sprites(), callback=get_sprites)
        return hud

    def run(self):
        self.running = True
        while self.running:
            self.frames += 1
            self.update_timer()
            self.handle_events()
            self.update(self.dt)
            self.draw()

    def switch_mode(self):
        self.mode = self.MODE_COMPOSITE if self.mode == self.MODE_SPRITES else self.MODE_SPRITES
        if self.mode == self.MODE_COMPOSITE:
            self.composite_buffer.drain()

    def update(self, dt):
        if self.mode == self.MODE_COMPOSITE:
            self.update_composite(dt)
        else:
            self.update_sprites(dt)

    def update_timer(self):
        self.dt = self.clock.tick(self.max_fps) / 1000.0
        self.timer -= self.dt
        if self.timer <= 0.0:
            self.timer = self.timer_reset
            mode = ('Composite', 'Sprites')
            ground = len(self.sprites[self.BACKGROUND])
            rocks = len(self.sprites[self.ON_GROUND])
            trees = len(self.sprites[self.OVERHEAD]) * 2
            composites = len(self.on_screen)
            things = 'Ground {}, Rocks {}, Trees+Shadows {}, Composites {}'.format(ground, rocks, trees, composites)
            pygame.display.set_caption('TAB: {}> | {} | FPS {}'.format(
                mode[self.mode], things, int(self.clock.get_fps())))
            self.hud.update()
            self.composite_buffer.prune(30)
            self.num_dirty = 0
            self.num_tiles_rendered = 0
            self.num_sprites_rendered = 0
            self.num_new = 0
            self.frames = 0

    def update_composite(self, dt):
        # update people, spatial hash, and composite buffer
        dirty_rects = []
        for ppl in self.sprites[self.ABOVE_GROUND]:
            dirty = False
            if ppl.dx or ppl.dy:
                dirty = True
                dirty_rects.append(tuple(ppl.rect))
            ppl.update(dt)
            if dirty:
                dirty_rects.append(tuple(ppl.rect))
        self.composite_src_layers[self.ABOVE_GROUND].addlist(self.sprites[self.ABOVE_GROUND])
        self.composite_buffer.set_dirty(dirty_rects)
        self.num_dirty += self.composite_buffer.num_dirty
        self.num_sprites_rendered += self.composite_buffer.num_sprites_rendered

        # update camera
        self.camera.center = self.player.rect.center
        self.camera.clamp_ip(self.world_rect)

        # get on-screen sprites (composite tiles)
        # self.on_screen = self.composite_buffer.get_tiles(self.camera, self.composite_hash)
        self.on_screen = self.composite_buffer.get_tiles(self.camera, self.sprite_getters)
        self.num_tiles_rendered += self.composite_buffer.num_tiles_rendered
        self.num_new += self.composite_buffer.num_new

    def update_sprites(self, dt):
        # update people and spatial hash
        for ppl in self.sprites[self.ABOVE_GROUND]:
            ppl.update(dt)
        self.composite_src_layers[self.ABOVE_GROUND].addlist(self.sprites[self.ABOVE_GROUND])

        # update camera
        self.camera.center = self.player.rect.center
        self.camera.clamp_ip(self.world_rect)

        # get on-screen sprites
        self.on_screen = []
        for sprites in self.composite_src_layers:
            on_screen = sorted(sprites.intersect_entities(self.camera), key=Sprite.sort_key)
            self.on_screen.extend(on_screen)

    def draw(self):
        self.screen.fill(self.clear_color)
        screen__blit = self.screen.blit
        cx, cy = self.camera.topleft
        for s in self.on_screen:
            screen__blit(s.image, s.rect.move(-cx, -cy))
        self.hud.draw(self.screen)
        pygame.display.flip()

    def handle_events(self):
        for e in pygame.event.get():
            if e.type == KEYDOWN:
                if e.key == K_UP:
                    self.player.dy -= 1
                elif e.key == K_DOWN:
                    self.player.dy += 1
                elif e.key == K_LEFT:
                    self.player.dx -= 1
                elif e.key == K_RIGHT:
                    self.player.dx += 1
                elif e.key == K_ESCAPE:
                    self.running = False
                elif e.key == K_TAB:
                    self.switch_mode()
                elif e.key == K_SPACE:
                    self.make_sprites()
            elif e.type == KEYUP:
                if e.key == K_UP:
                    self.player.dy += 1
                elif e.key == K_DOWN:
                    self.player.dy -= 1
                elif e.key == K_LEFT:
                    self.player.dx += 1
                elif e.key == K_RIGHT:
                    self.player.dx -= 1
            elif e.type == QUIT:
                self.running = False


if __name__ == '__main__':
    def main():
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        pygame.init()
        pygame.display.set_mode(RESOLUTION)
        Game().run()

    if False:
        cProfile.runctx('main()', globals(), locals(), 'prof.dat')
        p = pstats.Stats('prof.dat')
        p.sort_stats('time').print_stats()
    else:
        main()
