"""demo-03.py - 16x16 ground tiles, large map, scrolling

Background and mobile sprites use interpolation for smooth motion.

Two layers for the ground graphics.
One layer for the mobile graphics.
Two layers for the overhead graphics.

Four layers are combined into two composite buffers. The mobile layer is not composited.

2 composite buffers (one for ground, one for overhead)
5 sprite layers
5 spatial hashes
3 on-screen iists
"""
import cProfile
import os
import pstats
import random
import sys

import pygame
from pygame import Color, KEYDOWN, KEYUP, QUIT, K_i, K_UP, K_DOWN, K_LEFT, K_RIGHT, K_TAB, K_SPACE, K_ESCAPE

sys.path.insert(0, 'gamelib')

import clock
from compositebuffer import CompositeBuffer
from spatialhash import SpatialHash
import hudlight
import hudthemes


if sys.version_info[0] == 3:
    xrange = range

TIME_STEP = 1.0 / 30.0          # sim time, dt
MAX_FPS = 256                   # max frames rendered per second
SCREEN_RESOLUTION = 800, 600    # pixels
WORLD_SIZE = 4000, 4000         # pixels
WORLD_RECT = -WORLD_SIZE[0] // 2, -WORLD_SIZE[1] // 2, WORLD_SIZE[0], WORLD_SIZE[1]
WORLD_BOUNDS = -WORLD_SIZE[0] // 2, -WORLD_SIZE[1] // 2, WORLD_SIZE[0] // 2, WORLD_SIZE[1] // 2


def make_overhead_surf(surf, size):
    if not surf:
        surf = pygame.Surface(size, flags=pygame.SRCALPHA)
    surf.fill((0, 0, 0, 0), special_flags=pygame.BLEND_RGBA_MIN)
    return surf


class Sprite(object):
    layer = 0

    def __init__(self, image, rect, x, y, world_rect):
        self.image = image
        self.rect = rect
        self.x = x
        self.y = y
        self.world_rect = world_rect

    @staticmethod
    def sort_key(sprite):
        r = sprite.rect
        # rect.bottom, rect.x
        return sprite.layer, r[1] + r[3], r[0]

    def update(self, dt):
        pass


class Dirt(Sprite):
    layer = 0


class Rock(Sprite):
    layer = 1


class TreeShade(Sprite):
    layer = 0


class Tree(Sprite):
    layer = 1


class Player(Sprite):
    """keyboard directed sprite with interpolator"""
    layer = 1

    def __init__(self, image, rect, x, y, world_rect):
        super(Player, self).__init__(image, rect, x, y, world_rect)
        self.dx = 0
        self.dy = 0
        self.speed = 100
        self.interpolator = Interpolator(self.x, self.y)

    def update(self, dt):
        # move and stop at or slide along world edge
        if self.dx or self.dy:
            x = self.x + self.dx * self.speed * dt
            y = self.y + self.dy * self.speed * dt
            self.rect.center = x, y
            rect = self.rect.clamp(self.world_rect)
            if rect == self.rect:
                self.x, self.y = x, y
            else:
                x = x if rect.x == self.rect.x else rect.centerx
                y = y if rect.y == self.rect.y else rect.centery
                self.rect.center = x, y
                self.x, self.y = x, y
        self.interpolator.update(self.x, self.y)


class LadyFriend(Player):
    """steadily moving sprite with interpolator"""
    layer = 0

    def __init__(self, image, rect, x, y, world_rect):
        super(LadyFriend, self).__init__(image, rect, x, y, world_rect)
        self.dx = random.choice((-1, 1))
        self.dy = random.choice((-1, 1))

    def update(self, dt):
        # move and bounce off world edge
        x = self.x
        y = self.y
        new_x = x + self.dx * self.speed * dt
        new_y = y + self.dy * self.speed * dt
        self.rect.center = new_x, new_y
        if not self.world_rect.contains(self.rect):
            test_rect = self.rect.clamp(self.world_rect)
            if test_rect.x != self.rect.x:
                self.dx = -self.dx
            if test_rect.y != self.rect.y:
                self.dy = -self.dy
        self.x, self.y = new_x, new_y
        self.interpolator.update(self.x, self.y)


class Camera(object):
    """camera with interpolator"""
    def __init__(self, rect, world_rect):
        self.rect = rect
        self.world_rect = world_rect
        self.x = 0
        self.y = 0
        self.interpolator = Interpolator(0, 0)
        self.interpolate = self.interpolator.interpolate

    def _from_value(self):
        x, y = self.interpolator.from_value
        return x - self.rect.w // 2, y - self.rect.h // 2
    from_value = property(_from_value)

    def _to_value(self):
        x, y = self.interpolator.to_value
        return x - self.rect.w // 2, y - self.rect.h // 2
    to_value = property(_to_value)

    def set(self, x, y):
        self.rect.center = x, y
        self.x, self.y = x, y
        self.interpolator.set(x, y)

    def update(self, x, y):
        self__rect = self.rect
        self__rect.center = x, y
        rect = self__rect.clamp(self.world_rect)
        if rect == self__rect:
            self.x, self.y = x, y
            self.interpolator.update(x, y)
        else:
            x = x if rect.x == self__rect.x else rect.centerx
            y = y if rect.y == self__rect.y else rect.centery
            self.rect.center = x, y
            self.x, self.y = x, y
            self.interpolator.update(x, y)


class Interpolator(object):
    def __init__(self, *args):
        self._len = len(args)
        self.to_value = list(args)
        self.from_value = list(args)

    def set(self, *args):
        if len(args) != self._len:
            raise ValueError('length {} expected, got {}'.format(self._len, len(args)))
        self.from_value[:] = args
        self.to_value[:] = args

    def update(self, *args):
        if len(args) != self._len:
            raise ValueError('length {} expected, got {}'.format(self._len, len(args)))
        self.from_value[:] = self.to_value
        self.to_value[:] = args

    def interpolate(self, interp, reverse=False):
        if reverse is False:
            inter_values = [(t - f) * interp for t, f in zip(self.to_value, self.from_value)]
        else:
            revinterp = 1.0 - interp
            inter_values = [(t - f) * revinterp for t, f in zip(self.to_value, self.from_value)]
        return inter_values


class Game(object):
    # Tunable
    BACKGROUND_TILE_SIZE = 16   # background tiles will be square of this size
    NUM_ROCKS = 800             # num rocks!
    NUM_TREES = 500             # num trees and shadows!
    NUM_LADY_FRIENDS = 50       # num lady friends running away
    COMPOSITE_TILE_SIZE = 64    # size for CompositeBuffer and SpatialHash
    HUD_PERIOD = 0.5            # seconds between window caption and HUD updates

    MODE_COMPOSITE, MODE_SPRITES = 0, 1

    def __init__(self):
        self.screen = pygame.display.get_surface()
        self.screen_rect = self.screen.get_rect()
        self.clear_color = Color('black')

        self.dt = TIME_STEP
        self.draw_step = 1.0 / MAX_FPS
        self.hud_period = self.HUD_PERIOD
        self.clock = clock.FixedDriver(self.update, self.dt)
        self.draw_sched = self.clock.new_schedule(self.draw, self.draw_step, keep_history=True)
        self.clock.new_schedule(self.update_hud, self.hud_period)
        self.clock.new_schedule(self.update_caption, 0.5)
        self.clock.new_schedule(self.update_composite_buffers, 10.0)

        self.world_rect = pygame.Rect(WORLD_RECT)
        self.player = None
        self.camera = Camera(pygame.Rect(self.screen_rect), self.world_rect)

        # Composite buffers
        # 0: ground
        # 1: mobile, aka above ground
        # 2: overhead
        self.composite_buffers = [
            CompositeBuffer(tile_size=self.COMPOSITE_TILE_SIZE, sort_key=Sprite.sort_key),
            None,
            CompositeBuffer(tile_size=self.COMPOSITE_TILE_SIZE * 3 // 2, sort_key=Sprite.sort_key,
                            make_surf=make_overhead_surf),
        ]
        # Maintain a couple of nested structures for speed:
        # sprites for update and rendering; mode layers for compositing
        self.sprites = dict(
            ground=[],
            on_ground=[],
            above_ground=[],
            shadows=[],
            overhead=[],
        )
        self.space = dict(
            ground=SpatialHash(self.COMPOSITE_TILE_SIZE),
            on_ground=SpatialHash(self.COMPOSITE_TILE_SIZE),
            above_ground=SpatialHash(self.COMPOSITE_TILE_SIZE),
            shadows=SpatialHash(self.COMPOSITE_TILE_SIZE),
            overhead=SpatialHash(self.COMPOSITE_TILE_SIZE),
        )
        self.sprite_getters = (
            [self.space[name].intersect_entities for name in ('ground', 'on_ground')],
            [self.space[name].intersect_entities for name in ('above_ground',)],
            [self.space[name].intersect_entities for name in ('shadows', 'overhead')],
        )
        self.on_screen = [
            [],     # ground
            [],     # mobile
            [],     # overhead
        ]
        self.mode = self.MODE_SPRITES

        self.frames = 1
        self.num_dirty = 0
        self.num_tiles_rendered = 0
        self.num_sprites_rendered = 0
        self.num_new = 0

        self.make_sprites()
        self.hud = self.make_hud()

        self.caption_info = dict(
            mode=('Composite', 'Sprites'),
            ground=len(self.sprites['ground']),
            rocks=len(self.sprites['on_ground']),
            trees=len(self.sprites['shadows'] + self.sprites['overhead']),
        )

        self.camera.set(self.screen_rect.right, self.screen_rect.bottom)
        self.player.rect.center = self.camera.rect.center
        self.player.x, self.player.y = self.player.rect.center

        self.i = 0

        self.running = False

    def make_sprites(self):
        # Clear everything to make this reentrant.
        for sprites in self.sprites.values():
            del sprites[:]
        for space in self.space.values():
            space.clear()
        for on_screen in self.on_screen:
            del on_screen[:]
        self.composite_buffers[0].clear()
        # self.composite_buffers[1] is None, we don't use it
        self.composite_buffers[2].clear()

        # Dirt - make the background tiles
        font = pygame.font.SysFont('sans', 11)
        size = self.BACKGROUND_TILE_SIZE
        label = False
        total_tiles = (self.world_rect.height + size) * (self.world_rect.width + size) // size
        n = 0

        image = pygame.Surface((size, size))
        image.fill(Color('tan1'))
        # pygame.draw.rect(image, Color('tan2'), image.get_rect(), 1)

        y_range = xrange(self.world_rect.y, self.world_rect.bottom + size, size)
        for x in xrange(self.world_rect.x, self.world_rect.right + size, size):
            n += 1
            pygame.display.set_caption('Making background tiles {} of {}'.format(n, total_tiles))
            for y in y_range:
                _image = image
                rect = _image.get_rect(topleft=(x, y))
                pos_text = '{},{}'.format(x, y)
                if label and font.size(pos_text)[0] < size:
                    _image = _image.copy()
                    pos_label = font.render(pos_text, True, Color('black'))
                    _image.blit(pos_label, (1, 1))
                s = Dirt(_image, rect, x, y, self.world_rect)
                self.sprites['ground'].append(s)
        self.space['ground'].addlist(self.sprites['ground'])

        # Rocks
        total_tiles = self.NUM_ROCKS
        n = 0
        images = {}
        for i in range(self.NUM_ROCKS):
            n += 1
            pygame.display.set_caption('Making rock tiles {} of {}'.format(n, total_tiles))
            x = random.randrange(self.world_rect.x, self.world_rect.right)
            y = random.randrange(self.world_rect.y, self.world_rect.bottom)
            size = random.randrange(15, 25)
            r = size // 2
            if size in images:
                _image = images[size]
            else:
                _image = pygame.Surface((size, size))
                _image.set_colorkey(Color('black'))
                pygame.draw.circle(_image, Color('grey50'), (r, r), r)
                pygame.draw.circle(_image, Color('grey30'), (r, r), r, 1)
                images[size] = _image
            rect = _image.get_rect(topleft=(x, y)).clamp(self.world_rect)
            s = Rock(_image, rect, x, y, self.world_rect)
            self.sprites['on_ground'].append(s)
        self.space['on_ground'].addlist(self.sprites['on_ground'])

        # Player
        size = 40
        r = size // 2
        ppls_info = (
            (Player, 1, 'skyblue2', 'skyblue4'),
            (LadyFriend, self.NUM_LADY_FRIENDS, 'pink2', 'pink4'),
        )
        for Ppl, count, color1, color2 in ppls_info:
            color1 = pygame.Color(color1)
            color2 = pygame.Color(color2)
            image = pygame.Surface((size, size))
            image.set_colorkey(Color('black'))
            pygame.draw.circle(image, color1, (r, r), r)
            pygame.draw.circle(image, color2, (r, r), r, 1)
            for i in xrange(count):
                if Ppl is Player:
                    x, y = WORLD_BOUNDS[0], WORLD_BOUNDS[1]
                else:
                    x = random.randrange(self.world_rect.x, self.world_rect.right)
                    y = random.randrange(self.world_rect.y, self.world_rect.bottom)
                rect = image.get_rect(topleft=(x, y)).clamp(self.world_rect)
                x = rect.x
                y = rect.y
                s = Ppl(image, rect, x, y, self.world_rect)
                if Ppl is Player:
                    self.player = s
                self.sprites['above_ground'].append(s)
        self.space['above_ground'].addlist(self.sprites['above_ground'])

        # Trees and shadows
        total_tiles = self.NUM_TREES
        n = 0
        images = {}
        for i in xrange(self.NUM_TREES):
            n += 1
            pygame.display.set_caption('Making trees tiles {} of {}'.format(n, total_tiles))
            x = random.randrange(self.world_rect.x, self.world_rect.right)
            y = random.randrange(self.world_rect.y, self.world_rect.bottom)
            size = random.randrange(54, 75)
            r = size // 2
            if size in images:
                s_image, sh_image = images[size]
            else:
                s_image = pygame.Surface((size, size))
                s_image.set_colorkey(Color('black'))
                pygame.draw.circle(s_image, Color('green2'), (r, r), r)
                pygame.draw.circle(s_image, Color('green4'), (r, r), r, 1)
                sh_image = s_image.copy().convert_alpha()
                pygame.draw.circle(sh_image, Color('grey20'), (r, r), r)
                pygame.draw.circle(sh_image, Color('black'), (r, r), r, 1)
                sh_image.fill((255, 255, 255, 32), None, pygame.BLEND_RGBA_MULT)
                images[size] = s_image, sh_image
            s_rect = s_image.get_rect(center=(x, y)).clamp(self.world_rect)
            sh_rect = s_rect.move(45, 30)
            s = Tree(s_image, s_rect, s_rect.x, s_rect.y, self.world_rect)
            sh = TreeShade(sh_image, sh_rect, sh_rect.x, sh_rect.y, self.world_rect)
            self.sprites['shadows'].append(sh)
            self.sprites['overhead'].append(s)
        self.space['shadows'].addlist(self.sprites['shadows'])
        self.space['overhead'].addlist(self.sprites['overhead'])

        # Sort the sprites so they render in the same order as the composite tiles.
        for i, sprites in enumerate(self.sprites.values()):
            sprites.sort(key=self.composite_buffers[0].sort_key)

    def make_hud(self):
        def get_player():
            return self.player.rect.center

        def get_on_screen():
            return sum(len(seq) for seq in self.on_screen)

        def get_tiles():
            return len(self.composite_buffers[0])

        def get_cache():
            bytesize = self.screen.get_bytesize()
            num_cached = self.composite_buffers[0].num_cached
            cache_size = float(self.composite_buffers[0].cache_size * bytesize) / 2 ** 20
            cache_max = float(self.composite_buffers[0].cache_max_px * bytesize) / 2 ** 20
            return num_cached, cache_size, cache_max

        def get_comp_stats():
            num_dirty = self.num_dirty
            num_tiles_rendered = self.num_tiles_rendered
            frames = self.frames
            return int(num_dirty / frames), int(num_tiles_rendered / frames)

        def get_sprites():
            num_new = self.num_new
            num_sprites_rendered = self.num_sprites_rendered
            frames = self.frames
            return num_new, int(num_sprites_rendered / frames)

        hudlight.set_font_template('data/font/%s')
        theme = hudthemes.HUD_THEMES['verabd'].copy()
        theme.update({'fontsize': 18, 'color': 'lightblue1', 'gcolor': 'lightblue2', 'ocolor': 'black', 'owidth': 0.75})
        hud = hudlight.HUD(**theme)
        hud.add('fps', 'FPS {:0.0f}', 0, callback=self.draw_sched.per_second)
        hud.add('world', 'World bounds ({}, {}) - ({}, {})', *WORLD_BOUNDS)
        hud.add('player', 'Player pos {}, {}', *get_player(), callback=get_player)
        hud.add('on_screen', 'On Screen {}', get_on_screen(), callback=get_on_screen)
        hud.add('comp tiles', 'Comp Tiles {}', get_tiles(), callback=get_tiles)
        hud.add('comp cache', 'Comp Cache {} tiles, {:0.1f} / {:0.1f} MB', *get_cache(), callback=get_cache)
        hud.add('comp dirty', 'Comp Dirty {}/tick, Rendered {}/tick', *get_comp_stats(), callback=get_comp_stats)
        hud.add('comp sprites', 'Comp Sprites New {}, Rendered {}/tick', *get_sprites(), callback=get_sprites)
        return hud

    def run(self):
        self.running = True
        while self.running:
            self.handle_events()
            self.clock.tick()

    def switch_mode(self):
        self.mode = self.MODE_COMPOSITE if self.mode == self.MODE_SPRITES else self.MODE_SPRITES
        if self.mode == self.MODE_COMPOSITE:
            self.composite_buffers[0].drain()
            self.composite_buffers[2].drain()

    def update(self, dt):
        self.frames += 1
        if self.mode == self.MODE_COMPOSITE:
            self.update_composite(dt)
        else:
            self.update_sprites(dt)

    def update_composite_buffers(self, dt):
        self.composite_buffers[0].prune(20)
        # self.composite_buffers[0] is None, we don't use it
        self.composite_buffers[2].prune(20)

    def update_hud(self, dt):
        self.hud.update()
        self.num_dirty = 0
        self.num_tiles_rendered = 0
        self.num_sprites_rendered = 0
        self.num_new = 0
        self.frames = 0

    def update_caption(self, sched):
        on_screen = len(self.on_screen)
        ci = self.caption_info
        things = 'Ground {}, Rocks {}, Trees+Shadows {}, Rendering {}'.format(
            ci['ground'], ci['rocks'], ci['trees'], on_screen)
        pygame.display.set_caption('TAB: {}> | {} | FPS {}'.format(
            ci['mode'][self.mode], things, self.draw_sched.per_second()))

    def update_composite(self, dt):
        # update people, spatial hash, and composite buffer
        dirty_rects = []
        for ppl in self.sprites['above_ground']:
            dirty = False
            if ppl.dx or ppl.dy:
                dirty = True
                dirty_rects.append(tuple(ppl.rect))
            ppl.update(dt)
            if dirty:
                dirty_rects.append(tuple(ppl.rect))
        self.space['above_ground'].addlist(self.sprites['above_ground'])
        # only composite layer 0 needs to be marked dirty
        self.composite_buffers[0].set_dirty(dirty_rects)
        self.num_dirty += self.composite_buffers[0].num_dirty
        self.num_sprites_rendered += self.composite_buffers[0].num_sprites_rendered

        # update camera
        self.camera.update(self.player.x, self.player.y)

        # get on-screen sprites (composite tiles for ground and overhead, mobile sprites for middle layer)
        self.on_screen = [[], [], []]
        inflate = self.player.speed * self.dt * 2.0
        camera = self.camera.rect.inflate(inflate, inflate)
        # Note: we do not composite the mobile layer
        self.on_screen[0].extend(self.composite_buffers[0].get_tiles(camera, self.sprite_getters[0]))
        self.on_screen[1].extend(sorted(self.sprite_getters[1][0](camera), key=Sprite.sort_key))
        self.on_screen[2].extend(self.composite_buffers[2].get_tiles(camera, self.sprite_getters[2]))
        self.num_tiles_rendered += self.composite_buffers[0].num_tiles_rendered
        self.num_new += self.composite_buffers[0].num_new

    def update_sprites(self, dt):
        # update people and spatial hash
        for ppl in self.sprites['above_ground']:
            ppl.update(dt)
        self.space['above_ground'].addlist(self.sprites['above_ground'])

        # update camera
        self.camera.update(self.player.x, self.player.y)

        # get on-screen sprites
        self.on_screen = [[], [], []]
        inflate = self.player.speed * self.dt * 2.0
        camera = self.camera.rect.inflate(inflate, inflate)
        for i, name in ((0, 'ground'), (0, 'on_ground'), (1, 'above_ground'), (2, 'shadows'), (2, 'overhead')):
            self.on_screen[i].extend(sorted(
                self.space[name].intersect_entities(camera), key=Sprite.sort_key))

    def draw(self, sched):
        self.screen.fill(self.clear_color)
        self__camera = self.camera
        fm = self__camera.from_value
        to = self__camera.to_value
        if fm == to:
            self.draw_screen_stationary(sched.interp)
        else:
            self.draw_screen_moving(sched.interp)
        self.hud.draw(self.screen)
        pygame.display.flip()

    def draw_screen_stationary(self, interp):
        """calculate the least visually jarring experience for a stationary screen with mobile sprites"""
        screen__blit = self.screen.blit
        self__camera = self.camera
        camx, camy = self__camera.to_value
        for i, layer in enumerate(self.on_screen):
            if i == 1:
                # mobile sprite layer: do camera translation and local interpolation
                for s in layer:
                    ix, iy = s.interpolator.interpolate(interp)
                    screen__blit(s.image, s.image.get_rect(center=(s.x - camx + ix, s.y - camy + iy)))
            else:
                # static tile layer: do global translation only
                for s in layer:
                    screen__blit(s.image, s.rect.move(-camx, -camy))
                    # Uncomment this to see red outline for tree and shadow tiles
                    # if i == 2:
                    #     pygame.draw.rect(self.screen, Color('red'), s.rect.move(-camx, -camy), 1)

    def draw_screen_moving(self, interp):
        """calculate the least visually jarring experience for a moving screen with mobile sprites"""
        screen__blit = self.screen.blit
        self__camera = self.camera
        # polar coordinates and interpolated distance from the "from coordinates"
        fmx, fmy = self__camera.from_value
        tox, toy = self__camera.to_value
        ix, iy = self__camera.interpolate(interp)
        for i, layer in enumerate(self.on_screen):
            if i == 1:
                # mobile sprite layer: do camera translation
                for s in layer:
                    screen__blit(s.image, s.image.get_rect(center=(s.x - tox, s.y - toy)))
            else:
                # static tile layer: do camera translation and interpolation
                for s in layer:
                    screen__blit(s.image, s.rect.move(-fmx - ix, -fmy - iy))

    def handle_events(self):
        for e in pygame.event.get():
            if e.type == KEYDOWN:
                if e.key == K_UP:
                    self.player.dy -= 1
                elif e.key == K_DOWN:
                    self.player.dy += 1
                elif e.key == K_LEFT:
                    self.player.dx -= 1
                elif e.key == K_RIGHT:
                    self.player.dx += 1
                elif e.key == K_ESCAPE:
                    self.running = False
                elif e.key == K_TAB:
                    self.switch_mode()
                elif e.key == K_SPACE:
                    self.make_sprites()
                elif e.key == K_i:
                    self.i = (self.i + 1) % 8
            elif e.type == KEYUP:
                if e.key == K_UP:
                    self.player.dy += 1
                elif e.key == K_DOWN:
                    self.player.dy -= 1
                elif e.key == K_LEFT:
                    self.player.dx += 1
                elif e.key == K_RIGHT:
                    self.player.dx -= 1
            elif e.type == QUIT:
                self.running = False


if __name__ == '__main__':
    def main():
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        pygame.init()
        pygame.display.set_mode(SCREEN_RESOLUTION)
        Game().run()

    if False:
        cProfile.runctx('main()', globals(), locals(), 'prof.dat')
        p = pstats.Stats('prof.dat')
        p.sort_stats('time').print_stats()
    else:
        main()
