"""spatialhash.py - High performance spatial hash for spatial
partitioning and fast collision detection.

Objects (other than geometry.LineGeometry) must have a pygame Rect attribute.
Optionally, objects may have a collided static method attribute for lower-level
collision detection (see the gummworld2.geometry module).

Objects that are outside the world bounding rect are ignored by add().

As of v0.5.0 SpatialHash will sort objects returned by most methods that return
a list of objects. See the methods' docstrings. In order to have the objects
sorted set the SpatialHash's sort_key method. This is the same key that is
passed to sort(key=func) and sorted(key=func), e.g.:
key=lambda obj: obj.rect.bottom.

This module is derived from the article and source code written by Conkerjo at
http://conkerjo.wordpress.com/2009/06/13/spatial-hashing-implementation-for-fast-2d-collisions/.
"""

from collections import defaultdict

from pygame import Rect

import geometry


__all__ = ['SpatialHash']


class SpatialHash(object):
    def __init__(self, cell_size=50):
        self.cell_size = int(cell_size)
        # buckets = {(x, y): set(entity, ...)}
        self.buckets = defaultdict(set)
        # cell_ids = {entity: [(x, y), ...]}
        self.cell_ids = defaultdict(set)

        self._temp_rect = Rect(0, 0, 1, 1)

        self.sort_key = None
        self.dirty = False
        self.sorted_entities = []

        self.coll_tests = 0

    @property
    def entities(self):
        """Return the entire list of entities.
        
        This method honors self.sort_key.
        """
        if self.sort_key:
            if self.dirty:
                self.sorted_entities[:] = sorted(self.cell_ids, key=self.sort_key)
                self.dirty = False
            return self.sorted_entities[:]
        else:
            return self.cell_ids.keys()
    
    def add(self, entity):
        """Add or re-add obj. Return True if in bounds, else return False.
        
        If this method returns False then the object is completely out of
        bounds and cannot be stored in this space.
        
        Note that when obj changes its position, you must add it again so that
        its cell membership is updated. This method first removes the object if
        it is already in the spatial hash.
        """
        self.dirty = True
        self.addlist((entity, ))

    def addlist(self, entities):
        self.dirty = True
        self__cells = self.buckets
        self__cell_size = self.cell_size
        self__entity_to_cells = self.cell_ids
        self__remove = self.remove
        self__get_rect_for_line = self._get_rect_for_line
        for e in entities:
            try:
                rect = e.rect
            except AttributeError:
                rect = self__get_rect_for_line(e)
            x = rect[0]
            y = rect[1]
            w = rect[2]
            h = rect[3]
            if e in self__entity_to_cells:
                self__remove(e)

            cx_min = x // self__cell_size
            cx_max = (x + w) // self__cell_size
            cy_min = y // self__cell_size
            cy_max = (y + h) // self__cell_size

            y_range = range(cy_min, cy_max + 1)
            self__entity_to_cells_e__add = self__entity_to_cells[e].add
            for _x in range(cx_min, cx_max + 1):
                for _y in y_range:
                    _cell_id = _x, _y
                    self__cells[_cell_id].add(e)
                    self__entity_to_cells_e__add(_cell_id)

    def remove(self, entity):
        self.removelist([entity])

    def removelist(self, entities):
        self__buckets = self.buckets
        self__entity_to_cells = self.cell_ids
        self__entity_to_cells__pop = self__entity_to_cells.pop
        for e in entities:
            if e in self__entity_to_cells:
                for cell_id in self__entity_to_cells__pop(e):
                    self__buckets[cell_id].remove(e)
                    # remove empty bucket
                    if not self__buckets[cell_id]:
                        self__buckets.pop(cell_id)

    def get_nearby_entities(self, entity):
        """Return a list of entities that share the same cells as entity.
        
        This method honors self.sort_key.
        """
        nearby_ents = []
        extend = nearby_ents.extend
        try:
            cell_ids = self.intersect_cell_ids(entity.rect)
        except AttributeError:
            rect = self._get_rect_for_line(entity)
            cell_ids = self.intersect_cell_ids(rect)
        buckets = self.buckets
        for cell_id in cell_ids:
            if cell_id in buckets:
                extend(buckets[cell_id])
        if self.sort_key:
            nearby_ents.sort(key=self.sort_key)
        return nearby_ents
    
    def get_cell(self, cell_id):
        """Return the cell stored at bucket index cell_id.
        
        The returned cell is a list of entities. None is returned if a cell does
        not exist for cell_id.
        """
        if cell_id in self.buckets:
            return self.buckets[cell_id]
        else:
            return []
    
    def index(self, cell):
        """Return the bucket index of cell.
        
        Returns None if cell does not exist in buckets.
        
        Note that SpatialHash.buckets.index(cell) does *NOT* work because
        list.index() tests equality, not identity.
        """
        for i, c in self.buckets.items():
            if c is cell:
                return i
    
    def index_at(self, x, y):
        """Return the cell_id of the cell that contains point (x,y).
        
        None is returned if point (x,y) is not in bounds.
        """
        cell_size = self.cell_size
        idx = x // cell_size, y // cell_size
        return idx
    
    def intersect_cell_ids(self, rect):
        """Return list of cell ids that intersect rect.
        
        The return value is a list of int. Each int is a key for self.buckets. Note: the values are NOT checked for
        presence in buckets; so if self.buckets[i] is accessed directly, empty buckets may be accidentally created.
        
        :param rect: pygame.Rect; bounding rect
        :return: list of int
        """
        # Not pretty, but these ugly optimizations shave 50% off run-time
        # versus function calls and attributes. This method is called a lot.
        
        cell_size = self.cell_size

        left = rect[0] // cell_size
        top = rect[1] // cell_size
        right = (rect[0] + rect[2]) // cell_size
        bottom = (rect[1] + rect[3]) // cell_size

        cell_ids = []
        cell_ids_append = cell_ids.append
        for y in range(top, bottom + 1):
            for x in range(left, right + 1):
                cell_ids_append((x, y))

        return cell_ids
    
    def intersect_entities(self, rect):
        """Return list of entities whose rects intersect rect.
        
        This method honors self.sort_key.
        """
        result = {}
        colliderect = rect.colliderect
        rg = geometry.RectGeometry(*rect)
        buckets = self.buckets

        for cell_id in self.intersect_cell_ids(rect):
            if cell_id not in buckets:
                continue
            for o in buckets[cell_id]:
                try:
                    if colliderect(o.rect):
                        try:
                            if o.collided(o, rg, True):
                                result[o] = 1
                        except AttributeError:
                            result[o] = 1
                except AttributeError:
                    try:
                        if o.collided(o, rg, True):
                            result[o] = 1
                    except AttributeError:
                        pass
        result = list(result)
        if self.sort_key:
            result.sort(key=self.sort_key)
        return result
    
    def get_cell_pos(self, cell_id):
        """Return the world coordinates for topleft corner of cell.
        """
        x, y = cell_id
        cell_size = self.cell_size
        return x * cell_size, y * cell_size
    
    def collideany(self, entity):
        """Return True if entity collides with any other object, else False.
        """
        for other in self.get_nearby_entities(entity):
            if other is entity:
                continue
            try:
                if entity.rect.colliderect(other.rect):
                    try:
                        if entity.collided(entity, other, True):
                            return True
                    except AttributeError:
                        return True
            except AttributeError:
                try:
                    if entity.collided(entity, other, True):
                        return True
                except AttributeError:
                    pass
        return False
    
    def collide(self, entity):
        """Return list of entities that collide with entity.
        """
        collisions = []
        append = collisions.append
        tests = 0
        for other in self.get_nearby_entities(entity):
            if other is entity:
                continue
            tests += 1
            try:
                if entity.rect.colliderect(other.rect):
                    try:
                        if entity.collided(entity, other, True):
                            append(other)
                    except AttributeError:
                            append(other)
            except AttributeError:
                try:
                    if entity.collided(entity, other, True):
                        append(other)
                except AttributeError:
                    pass
        self.coll_tests = tests
        return collisions

    def collidedict(self, entities):
        """Return dict of entities that collide with entities.
        """
        result = defaultdict(set)
        buckets = self.buckets
        colls = set()
        result_pop = result.pop
        colls_add = colls.add
        colls_clear = colls.clear
        tests = 0

        cell_size = self.cell_size

        for e in entities:
            e__rect = e.rect

            left = e__rect[0] // cell_size
            top = e__rect[1] // cell_size
            right = (e__rect[0] + e__rect[2]) // cell_size
            bottom = (e__rect[1] + e__rect[3]) // cell_size

            # speed ups
            colliderect = getattr(e__rect, 'colliderect', None)
            e__collided = getattr(e, 'collided', None)
            result_e = result[e]
            result_e__remove = result_e.remove
            result_e__update = result_e.update
            x_range = range(left, right + 1)
            for y in range(top, bottom + 1):
                for x in x_range:
                    cell_id = x, y
                    if cell_id not in buckets:
                        continue
                    for o in buckets[cell_id]:
                        tests += 1
                        hit = False
                        if colliderect:
                            if colliderect(o.rect):
                                if e__collided:
                                    hit = e__collided(e, o, True)
                                else:
                                    hit = True
                        elif e__collided:
                            hit = e__collided(e, o, True)
                        if hit:
                            if o in result:
                                result[o].add(e)
                            else:
                                colls_add(o)
            if colls:
                result_e__update(colls)
                colls_clear()
            try:
                result_e__remove(e)
                if not result_e:
                    result_pop(e)
            except KeyError:
                pass

        self.coll_tests = tests
        return result

    def collidealldict(self, rect=None):
        """Return dict of all collisions.
        
        If rect is specified, only the cells that intersect rect will be
        checked.
        
        The contents of the returned dict are: {entity: [entity, ...}
        """
        result = defaultdict(set)
        self.coll_tests = 0
        if rect:
            buckets = self.buckets
            cells = [buckets[i] for i in self.intersect_cell_ids(rect)]
        else:
            cells = self.buckets.values()

        tests = 0
        for cell in cells:
            cell = list(cell)
            for i, e in enumerate(cell):
                e__rect = e.rect
                colls_e = []
                colls_e__append = colls_e.append
                try:
                    colliderect = e__rect.colliderect
                except AttributeError:
                    colliderect = None
                try:
                    e__collided = e.collided
                except AttributeError:
                    e__collided = None

                for o in cell[i+1:]:
                    if o is e:
                        continue
                    tests += 1
                    hit = False
                    try:
                        if colliderect(o.rect):
                            try:
                                if e__collided(e, o, True):
                                    hit = True
                            except TypeError:
                                hit = True
                    except TypeError:
                        try:
                            if e__collided(e, o, True):
                                hit = True
                        except TypeError:
                            pass
                    if hit:
                        if o in result:
                            result[o].add(e)
                        else:
                            colls_e__append(o)
                if colls_e:
                    result[e].update(colls_e)
                try:
                    if e in result:
                        result[e].remove(e)
                        if not result[e]:
                            result.pop(e)
                except KeyError:
                    pass

        self.coll_tests = tests
        return result

    def clear(self):
        """Clear all objects.
        """
        self.dirty = True
        self.buckets.clear()
        self.cell_ids.clear()
        self.sorted_entities = []

    def _get_rect_for_line(self, obj):
        """Lines don't have a rect attribute. Use self._temp_rect to fudge it.
        """
        points = obj.points
        rect = self._temp_rect
        x1, y1 = points[0]
        x2, y2 = points[1]
        if x1 > x2:
            t = x1
            x1 = x2
            x2 = t
        if y1 > y2:
            t = y1
            y1 = y2
            y2 = t
        rect.topleft = x1, y1
        rect.width = x2 - x1
        rect.height = y2 - y1
        return rect
    
    def __iter__(self):
        """This method honors self.sort_key."""
        if self.sort_key:
            if self.dirty:
                self.sorted_entities[:] = sorted(self.cell_ids, key=self.sort_key)
                self.dirty = False
            for obj in self.sorted_entities:
                yield obj
        else:
            for obj in self.cell_ids:
                yield obj
    
    def __contains__(self, obj):
        return obj in self.cell_ids
    
    def __len__(self):
        return len(self.cell_ids)
    
    def __str__(self):
        return '<{}({}) at {}>'.format(self.__class__.__name__, str(self.cell_size), id(self))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, str(self.cell_size))
